{
    This file is part of the Free Pascal Free Component Library.

    Unicode name file access and utility functions

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    The unit is based on FPC RTL SysUtils
    
 **********************************************************************}
unit
  ufiles;

interface

uses 
  {$ifdef MSWindows}
  Windows,
  {$endif}
  RtlConsts, Classes, SysUtils;
  
Type
  TSearchRecU = Record
    Time : Longint;
    Size : Int64;
    Attr : Longint;
    Name : UnicodeString;
    ExcludeAttr : LongInt;
    {$ifdef unix}
    FindHandle : Pointer;
    //mode?
    {$else}
    FindHandle : THandle;
    {$endif}
  end;

// 
function FindFirst (const Path : UnicodeString; Attr : Longint; out Rslt : TSearchRecU) : Longint;
function FindNext (var Rslt : TSearchRecU) : Longint;
procedure FindClose (var F : TSearchrecU);

function FileGetAttr (const FileName : UnicodeString) : Longint;
function FileSetAttr (const Filename : UnicodeString; Attr: longint) : Longint;

function FileOpen (const FileName : UnicodeString; Mode : Integer) : THandle;
function FileCreate (const FileName : UnicodeString) : THandle;
function FileCreate (const FileName : UnicodeString; Mode : Integer) : THandle;

// 
function FileIsReadOnly(const FileName: UnicodeString): Boolean;
function FileExists (const FileName : UnicodeString) : Boolean;
function DirectoryExists (const Directory : UnicodeString) : Boolean;
      
// 
function ChangeFileExt(const FileName, Extension: UnicodeString): UnicodeString;
function ExtractFilePath(const FileName: UnicodeString): UnicodeString;
function ExtractFileDrive(const FileName: Unicodestring): UnicodeString;
function ExtractFileName(const FileName: Unicodestring): UnicodeString;
function ExtractFileExt(const FileName: Unicodestring): UnicodeString;
function ExtractFileDir(Const FileName : Unicodestring): UnicodeString;
function ExtractRelativepath (Const BaseName,DestNAme : UnicodeString): UnicodeString;
function IncludeTrailingPathDelimiter(Const Path : UnicodeString) : UnicodeString;
function IncludeTrailingBackslash(Const Path : UnicodeString) : UnicodeString;
function ExcludeTrailingBackslash(Const Path: Unicodestring): UnicodeString;
function ExcludeTrailingPathDelimiter(Const Path: Unicodestring): UnicodeString;
function IsPathDelimiter(Const Path: Unicodestring; Index: Integer): Boolean;
function SameFileName(const S1, S2: UnicodeString): Boolean;
function GetDirs (var DirName : UnicodeString; Var Dirs : array of PUnicodeChar) : Longint;

type
  
  { TFileStreamU }

  TFileStreamU = class(THandleStream)
  Private
    FFileName : UnicodeString;
  public
    constructor Create(const AFileName: UnicodeString; Mode: Word);
    constructor Create(const AFileName: UnicodeString; Mode: Word; Rights: Cardinal);
    destructor Destroy; override;
    property FileName : UnicodeString Read FFilename;
  end;

implementation

type
  TAnsiCharSet = set of AnsiChar;

function UCharInSet(const wch: UnicodeChar; const AnsiSet: TAnsiCharSet): Boolean; inline;
begin
  Result := (Word(wch)<=127) and (Char(byte(wch)) in AnsiSet);
end;

function ChangeFileExt(const FileName, Extension: UnicodeString): UnicodeString;
var
  i : longint;
  EndSep : TAnsiCharSet;
begin
  i := Length(FileName);
  EndSep:=AllowDirectorySeparators+AllowDriveSeparators+[ExtensionSeparator];
  while (I > 0) and not UCharInSet(FileName[I], EndSep) do
    Dec(I);
  if (I = 0) or (FileName[I] <> ExtensionSeparator) then
    I := Length(FileName)+1;
  Result := Copy(FileName, 1, I - 1) + Extension;
end;

function ExtractFilePath(const FileName: UnicodeString): UnicodeString;
var
  i : longint;
  EndSep : Set of Char;
begin
  i := Length(FileName);
  EndSep:=AllowDirectorySeparators+AllowDriveSeparators;
  while (i > 0) and not UCharInSet(FileName[i], EndSep) do
    Dec(i);
  If I>0 then
    Result := Copy(FileName, 1, i)
  else
    Result:='';
end;
function ExtractFileDir(Const FileName : Unicodestring): UnicodeString;
var
  i : longint;
  EndSep : Set of Char;
begin
  I := Length(FileName);
  EndSep:=AllowDirectorySeparators+AllowDriveSeparators;
  while (I > 0) and not UCharInSet(FileName[I], EndSep) do
    Dec(I);
  if (I > 1) and UCharInSet(FileName[I], AllowDirectorySeparators) and
     not (FileName[I - 1] in EndSep) then
    Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileDrive(const FileName: Unicodestring): UnicodeString;
var
  i,l: longint;
begin
  Result := '';
  l:=Length(FileName);
  if (l<2) then
    exit;
  If (FileName[2] in AllowDriveSeparators) then
    result:=Copy(FileName,1,2)
  else if UCharInSet(FileName[1], AllowDirectorySeparators) and
          UCharInSet(FileName[2], AllowDirectorySeparators) then
    begin
      i := 2;
	
	  { skip share }
      While (i<l) and Not UCharInSet(Filename[i+1], AllowDirectorySeparators) do
        inc(i);
      inc(i);
	
      While (i<l) and Not UCharInSet(Filename[i+1], AllowDirectorySeparators) do
        inc(i);
      Result:=Copy(FileName,1,i);
    end;
end;

function ExtractFileName(const FileName: Unicodestring): UnicodeString;
var
  i : longint;
  EndSep : Set of Char;
begin
  I := Length(FileName);
  EndSep:=AllowDirectorySeparators+AllowDriveSeparators;
  while (I > 0) and not UCharInSet(FileName[I], EndSep) do
    Dec(I);
  Result := Copy(FileName, I + 1, MaxInt);
end;

function ExtractFileExt(const FileName: Unicodestring): UnicodeString;
var
  i : longint;
  EndSep : Set of Char;
begin
  I := Length(FileName);
  EndSep:=AllowDirectorySeparators+AllowDriveSeparators+[ExtensionSeparator];
  while (I > 0) and not UCharInSet(FileName[I], EndSep) do
    Dec(I);
  if (I > 0) and (FileName[I] = ExtensionSeparator) then
    Result := Copy(FileName, I, MaxInt)
  else
    Result := '';
end;


function StrIcompU(p1, p2: PUnicodeChar): Integer;
var
  l1, l2 : Integer;
  ws1, ws2 : WideString;
begin
  l1:=0;
  while p1[l1]<>#0 do inc(l1);
  l2:=0;
  while p2[l1]<>#0 do inc(l2);
  Result:=l1-l2;
  if Result<>0 then begin
    if Result<0 then Result:=-1
    else Result:=1;
    Exit;
  end;
  
  if l1=0 then begin
    Result:=0; 
    Exit;
  end;
  SetLength(ws1, l1);
  SetLength(ws2, l1);
  Move(p1^, ws1[1], l1*sizeof(UnicodeChar));
  Move(p2^, ws2[1], l1*sizeof(UnicodeChar));
  Result:=WideCompareText(ws1, ws2) 
end;

function ExtractRelativepath (Const BaseName,DestName : UnicodeString): UnicodeString;
const
  MaxDirs = 129;
Var 
  Source, Dest : UnicodeString;
  Sc,Dc,I,J    : Longint;
  SD,DD        : Array[1..MaxDirs] of PWideChar;
const 
  OneLevelBack = '..'+DirectorySeparator;

begin
  If WideUpperCase(ExtractFileDrive(BaseName))<>WideUpperCase(ExtractFileDrive(DestName)) then begin
    Result:=DestName;
    exit;
  end;
  Source:=ExcludeTrailingPathDelimiter(ExtractFilePath(BaseName));
  Dest:=ExcludeTrailingPathDelimiter(ExtractFilePath(DestName));
  SC:=GetDirs (Source,SD);
  DC:=GetDirs (Dest,DD);
  I:=1;
  While (I<=DC) and (I<=SC) do
    begin
    If StrIcompU(DD[i],SD[i])=0 then // wrong. can't use PChar function!
      Inc(i)
    else
      Break;
    end;
  Result:='';
  For J:=I to SC do Result:=Result+OneLevelBack;
  For J:=I to DC do Result:=Result+DD[J]+DirectorySeparator;
  Result:=Result+ExtractFileName(DestNAme);
end;

function IncludeTrailingPathDelimiter(Const Path : UnicodeString) : UnicodeString;
Var
  l : Integer;
begin
  Result:=Path;
  l:=Length(Result);
  If (L=0) or not UCharInSet(Result[l], AllowDirectorySeparators) then
    Result:=Result+DirectorySeparator;
end;

function IncludeTrailingBackslash(Const Path : UnicodeString) : UnicodeString;
begin
  Result:=IncludeTrailingPathDelimiter(Path);
end;

function ExcludeTrailingBackslash(Const Path: Unicodestring): UnicodeString;
begin
  Result:=ExcludeTrailingPathDelimiter(Path);
end;

function ExcludeTrailingPathDelimiter(Const Path: Unicodestring): UnicodeString;
var
  L : Integer;
begin
  L:=Length(Path);
  If (L>0) and (Path[L] in AllowDirectorySeparators) then
    Dec(L);
  Result:=Copy(Path,1,L);
end;

function IsPathDelimiter(Const Path: Unicodestring; Index: Integer): Boolean;
begin
  Result:=(Index>0) and (Index<=Length(Path)) and UCharInSet(Path[Index], AllowDirectorySeparators);
end;

function SameFileName(const S1, S2: UnicodeString): Boolean;
begin
  If FileNameCaseSensitive then begin
    Result:=WideCompareText(S1,S2)=0; // Compare case sensitive
  end else begin
    // Compare case insensitive. No MBCS yet.
    Result:=WideCompareText(WideLowerCase(S1), WideLowerCase(S2)) =0;
  end;
end;

function GetDirs (var DirName : UnicodeString; Var Dirs : array of PUnicodeChar) : Longint;
Var 
  I : Longint;
begin
  I:=1;
  Result:=-1;
  While I<=Length(DirName) do
    begin
    If UCharInSet(DirName[i], AllowDirectorySeparators) and
       { avoid error in case last char=pathdelim }
       (length(dirname)>i) then
      begin
        DirName[i]:=#0;
        Inc(Result);
        Dirs[Result]:=@DirName[I+1];
      end;
    Inc(I);
    end;
  If Result>-1 then inc(Result);
end;


// Ansi version functions.
// The functions are using SysUtils file functions
// File names are converted from Unicode to current system Locale
// Can be used for UTF-8 (modern unixes) OSes, but should not be used for WinNT

function FileGetAttrAnsi (const FileName : UnicodeString) : Longint;
begin
  Result:=SysUtils.FileGetAttr(FileName);
end;

function FileSetAttrAnsi (const Filename : UnicodeString; Attr: longint) : Longint;
begin
  Result:=SysUtils.FileSetAttr(FileName, Attr);
end;

procedure SearchRecAnsiToUnicode(const src: TSearchRec; var dst: TSearchRecU);
begin
  with dst do begin
    Time := src.Time;
    Size := src.Size;
    Attr := src.Attr;
    Name := src.Name;
    FindHandle := src.FindHandle;
  end;
end;

procedure SearchRecUnicodeToAnsi(const src: TSearchRecU; var dst: TSearchRec);
begin
  // for additional structures
  FillChar(dst, sizeof(dst), 0); 
  with dst do begin
    Time := src.Time;
    Size := src.Size;
    Attr := src.Attr;
    Name := src.Name;
    FindHandle := src.FindHandle;
  end;
end;

function FindFirstAnsi (Const Path : UnicodeString; Attr : Longint; out Rslt : TSearchRecU) : Longint;
var
  sr : TSearchRec;
begin
  Result:=SysUtils.FindFirst(Path, Attr, sr);
  if Result=0 then SearchRecAnsiToUnicode(sr, Rslt)    
end;

function FindNextAnsi (var Rslt : TSearchRecU) : Longint;
var
  sr : TSearchRec;
begin
  SearchRecUnicodeToAnsi(Rslt, sr);
  Result:=SysUtils.FindNext(sr);
  SearchRecAnsiToUnicode(sr, Rslt);
end;

procedure FindCloseAnsi (var F : TSearchrecU);
var
  sr : TSearchRec;
begin
  SearchRecUnicodeToAnsi(F, sr);
  SysUtils.FindClose(sr);
end;

function FileOpenAnsi(const FileName : UnicodeString; Mode : Integer): THandle;
begin                                                                
  Result:=SysUtils.FileOpen(FileName, Mode);
end;

function FileCreateAnsi(const FileName : UnicodeString; Mode : Integer) : THandle;
begin
  Result:=SysUtils.FileCreate(FileName, Mode);  
end;

// Cross-platform part
// Do* functions are system dependable and can be reassigned, if necessary

var
  DoFindFirst : function  (const Path : UnicodeString; Attr : Longint; out Rslt : TSearchRecU) : Longint = @FindFirstAnsi;
  DoFindNext  : function  (var Rslt : TSearchRecU) : Longint = @FindNextAnsi;
  DoFindClose : procedure (var F : TSearchrecU) = @FindCloseAnsi;
  DoFileGetAttr : function (const FileName : UnicodeString) : Longint = @FileGetAttrAnsi;
  DoFileSetAttr : function (const Filename : UnicodeString; Attr: longint) : Longint = @FileSetAttrAnsi;
  DoFileOpen    : function (const FileName : UnicodeString; Mode : Integer) : THandle = @FileOpenAnsi;
  DoFileCreate  : function (const FileName : UnicodeString; Mode : Integer) : THandle = @FileCreateAnsi;
  
  
function FileGetAttr (const FileName : UnicodeString) : Longint;
begin
  Result:=DoFileGetAttr(FileName);
end;

function FileSetAttr (const Filename : UnicodeString; Attr: longint) : Longint;
begin
  Result:=DoFileSetAttr(FileName, Attr);
end;

function FileExists (const FileName : UnicodeString) : Boolean;
var
  Attr  : Integer;
begin
  Attr:=FileGetAttr(FileName);
  Result:=Attr<>-1;
  if Result then
    Result:=(Attr and faDirectory)=0;
end;

function DirectoryExists (const Directory : UnicodeString) : Boolean;
var
  Attr  : Integer;
begin
  Attr:=FileGetAttr(Directory);
  Result:=Attr<>-1;
  if Result then
    Result:=(Attr and faDirectory)<>0;
end;

function FindFirst (const Path : UnicodeString; Attr : Longint; out Rslt : TSearchRecU) : Longint;
begin
  Result:=DoFindFirst(Path, Attr, Rslt);
end;

function FindNext (var Rslt : TSearchRecU) : Longint;
begin
  Result:=DoFindNext(Rslt);
end;

function FileIsReadOnly(const FileName: UnicodeString): Boolean;
begin
  Result := (FileGetAttr(FileName) and faReadOnly) <> 0;
end;

procedure FindClose (var F : TSearchrecU);
begin               
  DoFindClose(f);
end;

function FileOpen (const FileName : UnicodeString; Mode : Integer) : THandle;
begin
 Result:=DoFileOpen(FileName, Mode);
end;

function FileCreate (const FileName : UnicodeString) : THandle;
begin
  Result:=DoFileCreate(FileName, fmCreate);
end;

function FileCreate (const FileName : UnicodeString; Mode : Integer) : THandle;
begin
  Result:=DoFileCreate(FileName, Mode);  
end;

// ---------------------------- Windows Specific ---------------------------- 
{$ifdef MSWindows}
var
  kernel32dll    : THandle = 0; 
  
  FindFirstFileU : function (lpFileName: LPWSTR; var lpFindFileData: TWIN32FindDataW): THandle; stdcall = nil;
  FindNextFileU  : function (hFindFile: THandle; var lpFindFileData: TWIN32FindDataW): BOOL; stdcall = nil;

  SetFileAttributesU : function (lpFileName:LPCWSTR; dwFileAttributes:DWORD):WINBOOL; stdcall = nil;
  GetFileAttributesU : function (lpFileName:LPCWSTR):DWORD; stdcall = nil;
  
  CreateFileU : function (lpFileName:LPCWSTR; dwDesiredAccess:DWORD; 
    dwShareMode:DWORD; lpSecurityAttributes:LPSECURITY_ATTRIBUTES; 
    dwCreationDisposition:DWORD;dwFlagsAndAttributes:DWORD; 
    hTemplateFile:HANDLE):HANDLE; stdcall = nil;


Function WinToDosTime (Var Wtime : TFileTime;var DTime:longint):longbool;
var
  lft : TFileTime;
begin
  WinToDosTime:=FileTimeToLocalFileTime(WTime,lft) and
                FileTimeToDosDateTime(lft,Longrec(Dtime).Hi,LongRec(DTIME).lo);
end;

Function FindMatch(var FindData: TWin32FindDataW; var f: TSearchRecU) : Longint;
begin
  { Find file with correct attribute }
  While (FindData.dwFileAttributes and cardinal(F.ExcludeAttr))<>0 do
   begin
     if not FindNextFileU (F.FindHandle, FindData) then
      begin
        Result:=GetLastError;
        exit;
      end;
   end;
  { Convert some attributes back }
  WinToDosTime(FindData.ftLastWriteTime,F.Time);
  f.size:=FindData.NFileSizeLow+(qword(maxdword)+1)*FindData.NFileSizeHigh;
  f.attr:=FindData.dwFileAttributes;
  f.Name:=FindData.cFileName;
  Result:=0;
end;

function FindFirstWin(const Path: UnicodeString; Attr: Longint; out Rslt: TSearchRecU): Longint;
var
  find  : TWIN32FINDDATAW;
begin
  Rslt.Name:=Path;
  Rslt.Attr:=attr;
  Rslt.ExcludeAttr:=(not Attr) and ($1e);

  Rslt.FindHandle:=FindFirstFileU( PWideChar(Path),find);
  If Rslt.FindHandle=Invalid_Handle_value then
  begin
    Result:=GetLastError;
    Exit;
  end;
  { Find file with correct attribute }
  Result:=FindMatch(find, Rslt);
end;

function FindNextWin(var Rslt: TSearchRecU): Longint;
var
  find  : TWIN32FINDDATAW;
begin
  if FindNextFileU(Rslt.FindHandle, find) then 
    Result := FindMatch(find, Rslt)
  else
    Result := GetLastError;
end;

Procedure FindCloseWin(Var F : TSearchRecU);
begin
  if F.FindHandle <> INVALID_HANDLE_VALUE then
    Windows.FindClose(F.FindHandle);
end;

function FileGetAttrWin(const FileName : UnicodeString) : Longint;
begin
  Result:=GetFileAttributesU(PWideChar(FileName));
end;

function FileSetAttrWin(const Filename : UnicodeString; Attr: longint) : Longint;
begin
  if SetFileAttributesU(PWideChar(FileName), Attr) then
    Result:=0
  else
    Result := GetLastError;
end;

function FileOpenWin(const FileName: UnicodeString; Mode : Integer): THandle;
const
  AccessMode: array[0..2] of Cardinal  = (
    GENERIC_READ,
    GENERIC_WRITE,
    GENERIC_READ or GENERIC_WRITE);
  ShareMode: array[0..4] of Integer = (
               0,
               0,
               FILE_SHARE_READ,
               FILE_SHARE_WRITE,
               FILE_SHARE_READ or FILE_SHARE_WRITE);
begin
  Result := CreateFileU(PWideChar(FileName), dword(AccessMode[Mode and 3]),
                       dword(ShareMode[(Mode and $F0) shr 4]), nil, OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL, 0);
  //if fail api return feInvalidHandle (INVALIDE_HANDLE=feInvalidHandle=-1)
end;

function FileCreateWin(const FileName: UnicodeString; Mode : Integer): THandle;
begin
  Result := CreateFileU(PWidechar(FileName), GENERIC_READ or GENERIC_WRITE,
              0, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
end;


procedure InitWindows;
begin
  kernel32dll:=LoadLibrary('kernel32.dll');
  if kernel32dll=0 then Exit;
  
  Pointer(FindFirstFileU) := GetProcAddress(kernel32dll, 'FindFirstFileW');
  Pointer(FindNextFileU) := GetProcAddress(kernel32dll, 'FindNextFileW');

  Pointer(SetFileAttributesU) := GetProcAddress(kernel32dll, 'SetFileAttributesW'); 
  Pointer(GetFileAttributesU) := GetProcAddress(kernel32dll, 'GetFileAttributesW'); 
  
  Pointer(CreateFileU) := GetProcAddress(kernel32dll, 'CreateFileW');
   
  if Assigned(FindFirstFileU) and Assigned(FindNextFileU) then begin
    DoFindFirst:=@FindFirstWin;   
    DoFindNext:=@FindNextWin;
    DoFindClose:=@FindCloseWin;
  end;
  
  if Assigned(SetFileAttributesU) and Assigned(GetFileAttributesU) then begin
    DoFileGetAttr:=@FileGetAttrWin;
    DoFileSetAttr:=@FileSetAttrWin;
  end;
  
  if Assigned(CreateFileU) then begin
    DoFileOpen:=@FileOpenWin;
    DoFileCreate:=@FileCreateWin;
  end;
  
end;

procedure ReleaseWindows;
begin
  if kernel32dll<>0 then FreeLibrary(kernel32dll);
end;


{$endif}
// ---------------------------- Windows Specific ---------------------------- 

{ TFileStreamU }

constructor TFileStreamU.Create(const AFileName: UnicodeString; Mode: Word); 
var
  hnd : THandle;
begin
  FFileName:=AFileName;
  If Mode=fmcreate then
    hnd:=FileCreate(AFileName)
  else
    hnd:=FileOpen(AFileName,Mode);

  If (THandle(hnd)=feInvalidHandle) then
    If Mode=fmcreate then
      raise EFCreateError.createfmt(SFCreateError,[AFileName])
    else
      raise EFOpenError.Createfmt(SFOpenError,[AFilename]);
  inherited Create(hnd);
end;

constructor TFileStreamU.Create(const AFileName: UnicodeString; Mode: Word; Rights: Cardinal); 
var
  hnd : THandle;
begin
  FFileName:=AFileName;
  If Mode=fmcreate then
    hnd:=FileCreate(AFileName,Rights)
  else
    hnd:=FileOpen(AFileName,Mode);

  If (THandle(hnd)=feInvalidHandle) then
    If Mode=fmcreate then
      raise EFCreateError.createfmt(SFCreateError,[AFileName])
    else
      raise EFOpenError.Createfmt(SFOpenError,[AFilename]);
  inherited Create(hnd);
end;

destructor TFileStreamU.Destroy;  
begin
  FileClose(Handle);
  inherited Destroy;  
end;

initialization
  {$ifdef MSWindows}
  InitWindows;
  {$endif}
  
finalization
  {$ifdef MSWindows}
  ReleaseWindows;
  {$endif}
  
end.
